// Copyright (C) 2000 Jean-Claude Wippler <jcw@equi4.com>
//
// Reflecting channel interface

//#include <tcl.h>
#include <tclInt.h>
#include <tclPort.h>
#include <tclIO.h>
#include <memory.h>

//typedef Tcl_Channel (TclOpenFileChannelProc_) (Tcl_Interp*,char*,char*,int);
extern "C" int TclOpenFileChannelInsertProc(TclOpenFileChannelProc_*);

#ifndef EINVAL
#define EINVAL 9
#endif

extern "C" DLLEXPORT int Rechan_Init(Tcl_Interp*);

///////////////////////////////////////////////////////////////////////////////

class ReflectingChannel
{
public:
    ReflectingChannel (Tcl_Interp*, Tcl_Obj*, int, const char*);
    ~ReflectingChannel ();

    Tcl_Channel _chan;
    int _validMask;
    int _watchMask;
    Tcl_Interp* _interp;
    Tcl_Obj* _context;
    Tcl_Obj* _seek;
    Tcl_Obj* _read;
    Tcl_Obj* _write;
    Tcl_Obj* _name;
};

ReflectingChannel::ReflectingChannel (Tcl_Interp* ip_, Tcl_Obj* context_,
                                      int mode_, const char* name_)
    : _validMask (mode_), _watchMask (0), _chan (0),
      _context (context_), _interp (ip_), _name (Tcl_NewStringObj(name_, -1))
{
        // support Tcl_GetIndexFromObj by keeping these objectified
    _seek = Tcl_NewStringObj("seek", -1);
    _read = Tcl_NewStringObj("read", -1);
    _write = Tcl_NewStringObj("write", -1);

    Tcl_IncrRefCount(_context);
    Tcl_IncrRefCount(_seek);
    Tcl_IncrRefCount(_read);
    Tcl_IncrRefCount(_write);
    Tcl_IncrRefCount(_name);
}

ReflectingChannel::~ReflectingChannel ()
{
    //printf("destroying reflchannel %s\n", Tcl_GetString(_name));
    Tcl_DecrRefCount(_context);
    Tcl_DecrRefCount(_seek);
    Tcl_DecrRefCount(_read);
    Tcl_DecrRefCount(_write);
    Tcl_DecrRefCount(_name);

    if (_chan != 0)
	Tcl_UnregisterChannel(_interp, _chan);
}

class RcEvent {
public:
    Tcl_Event _header;
    ReflectingChannel* chan;
};

static int
rcEventProc(Tcl_Event *evPtr, int flags)
{
    RcEvent *me = (RcEvent*) evPtr;

    if (!(flags & TCL_FILE_EVENTS))
        return 0;

    Tcl_NotifyChannel(me->chan->_chan, me->chan->_watchMask);
    return 1;
}

static void
rcCheckProc(ClientData cd_, int flags)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;

    if (!(flags & TCL_FILE_EVENTS))
        return;

    RcEvent *evPtr = (RcEvent*) Tcl_Alloc(sizeof (RcEvent));
    evPtr->_header.proc = rcEventProc;
    evPtr->chan = chan;

    Tcl_QueueEvent( (Tcl_Event*)evPtr, TCL_QUEUE_TAIL);
}

static void
rcSetupProc(ClientData cd_, int flags)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    Tcl_Time blockTime = { 0, 0};

    if (!(flags & TCL_FILE_EVENTS))
        return;

    if (chan->_watchMask)
        Tcl_SetMaxBlockTime(&blockTime);
}

static int
rcEventFilter( Tcl_Event *evPtr, ClientData cd_)
{
    RcEvent* me = (RcEvent*) evPtr;
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    return evPtr->proc == rcEventProc && me->chan == chan;
}

static int
rcClose (ClientData cd_, Tcl_Interp* interp)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    //printf("closing refl-channel %s\n", Tcl_GetString(chan->_name));

//  Tcl_DeleteEventSource(rcSetupProc, rcCheckProc, chan);
    Tcl_DeleteEvents(rcEventFilter, chan);
    delete chan;

    return TCL_OK;
}

static Tcl_Obj*
rcBuildCmdList(ReflectingChannel* chan_, Tcl_Obj* cmd_)
{
    Tcl_Obj* vec = Tcl_DuplicateObj(chan_->_context);
    Tcl_IncrRefCount(vec);
    if (Tcl_ListObjAppendElement(chan_->_interp, vec, cmd_) == TCL_ERROR
     || Tcl_ListObjAppendElement(chan_->_interp, vec, chan_->_name) == TCL_ERROR)
    {
        Tcl_DecrRefCount(vec);
        return 0;
    }
    return vec; // with refcount 1
}

static int
rcInput (ClientData cd_, char* buf, int toRead, int* errorCodePtr)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    Tcl_Interp* ip = chan->_interp;

    Tcl_Obj* cmd = rcBuildCmdList(chan, chan->_read);
    if (cmd != 0)
    {
	int n = -1;
        int e = Tcl_ListObjAppendElement(ip, cmd, Tcl_NewIntObj(toRead));
        if (e == TCL_OK)
        {
	    Tcl_SavedResult sr;
	    Tcl_SaveResult(ip, &sr);
            e = Tcl_EvalObjEx(ip, cmd, TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
            if (e == TCL_OK)
            {
                Tcl_Obj* r = Tcl_GetObjResult(ip);
                unsigned char* s = Tcl_GetByteArrayFromObj(r, &n);
                if (0 <= n && n <= toRead)
                    memcpy(buf, s, n);
		else
		    n = -1;
            }
	    Tcl_RestoreResult(ip, &sr);
        }
        Tcl_DecrRefCount(cmd);
	if (n >= 0)
	    return n;
    }

    *errorCodePtr = EINVAL;
    return -1;
}

static int
rcOutput (ClientData cd_, char* buf, int toWrite, int* errorCodePtr)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    Tcl_Interp* ip = chan->_interp;

    Tcl_Obj* cmd = rcBuildCmdList(chan, chan->_write);
    if (cmd != 0)
    {
        int e = Tcl_ListObjAppendElement(ip, cmd,
                        Tcl_NewByteArrayObj((unsigned char*) buf, toWrite));
        if (e == TCL_OK)
        {
	    Tcl_SavedResult sr;
	    Tcl_SaveResult(ip, &sr);
            e = Tcl_EvalObjEx(ip, cmd, TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
            if (e == TCL_OK)
            {
                Tcl_Obj* r = Tcl_GetObjResult(ip);
                int n;
                e = Tcl_GetIntFromObj(ip, r, &n);
                if (e == TCL_OK && 0 <= n && n <= toWrite) {
		    Tcl_RestoreResult(ip, &sr);
                    Tcl_DecrRefCount(cmd);
                    return n;
                }
            }
	    Tcl_RestoreResult(ip, &sr);
        }
        Tcl_DecrRefCount(cmd);
    }

    *errorCodePtr = EINVAL;
    return -1;
}

static int
rcSeek (ClientData cd_, long offset, int seekMode, int* errorCodePtr)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    Tcl_Interp* ip = chan->_interp;

    Tcl_Obj* cmd = rcBuildCmdList(chan, chan->_seek);
    if (cmd != 0)
    {
        int e = Tcl_ListObjAppendElement(ip, cmd, Tcl_NewLongObj(offset));
        if (e == TCL_OK)
        {
            int e = Tcl_ListObjAppendElement(ip, cmd, Tcl_NewIntObj(seekMode));
            if (e == TCL_OK)
            {
		Tcl_SavedResult sr;
		Tcl_SaveResult(ip, &sr);
                e = Tcl_EvalObjEx(ip, cmd, TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
                if (e == TCL_OK)
                {
                    Tcl_Obj* r = Tcl_GetObjResult(ip);
                    int n;
                    e = Tcl_GetIntFromObj(ip, r, &n);
                    if (e == TCL_OK) {
			Tcl_RestoreResult(ip, &sr);
                        Tcl_DecrRefCount(cmd);
                        return n;
                    }
                }
		Tcl_RestoreResult(ip, &sr);
            }
        }
        Tcl_DecrRefCount(cmd);
    }

    *errorCodePtr = EINVAL;
    return -1;
}

static void
rcWatchChannel (ClientData cd_, int mask)
{
    ReflectingChannel* chan = (ReflectingChannel*) cd_;
    Tcl_Time blockTime = { 0, 0 };

    /*
     * Since the file is always ready for events, we set the block time
     * to zero so we will poll.
     */

    chan->_watchMask = mask & chan->_validMask;
    if (chan->_watchMask) {
        Tcl_SetMaxBlockTime(&blockTime);
    }
}

static int
rcGetFile (ClientData cd_, int direction, ClientData* handlePtr)
{
    return TCL_ERROR;
}

static Tcl_ChannelType reChannelType = {
  "rechan",         /* Type name.                                    */
  0,                /* Set blocking/nonblocking behaviour. NULL'able */
  rcClose,          /* Close channel, clean instance data            */
  rcInput,          /* Handle read request                           */
  rcOutput,         /* Handle write request                          */
  rcSeek,           /* Move location of access point.      NULL'able */
  0,                /* Set options.                        NULL'able */
  0,                /* Get options.                        NULL'able */
  rcWatchChannel,   /* Initialize notifier                           */
  rcGetFile         /* Get OS handle from the channel.               */
};

///////////////////////////////////////////////////////////////////////////////

static int
RechanObjCmd(ClientData cd_, Tcl_Interp* ip_, int objc_, Tcl_Obj*const* objv_)
{
    if (objc_ != 3) {
        Tcl_WrongNumArgs(ip_, 1, objv_, "command mode");
        return TCL_ERROR;
    }

    int mode;
    if (Tcl_GetIntFromObj(ip_, objv_[2], &mode) == TCL_ERROR)
        return TCL_ERROR;

    static int mkChanSeq = 0;
    char buffer [20];
    sprintf(buffer, "rechan%d", ++mkChanSeq);

    ReflectingChannel* rc = new ReflectingChannel (ip_, objv_[1], mode, buffer);

    rc->_chan = Tcl_CreateChannel(&reChannelType, buffer, rc, mode);

//  Tcl_CreateEventSource(rcSetupProc, rcCheckProc, rc);

    Tcl_RegisterChannel(ip_, rc->_chan);

    Tcl_SetResult(ip_, buffer, TCL_VOLATILE);
    return TCL_OK;
}

///////////////////////////////////////////////////////////////////////////////

    static int recursive;

    	// one mutex is not optimal in the multi-threaded case, because the
	// "eval" in the code below will run while it is set, i.e. monolithic
	// in other words: this open will complete before switching threads
    TCL_DECLARE_MUTEX(icMutex)

    Tcl_Interp* someInterp;

static Tcl_Channel
VfsChanFileOpen(Tcl_Interp* ip_, char* fn_, char* mode_, int perm_)
{
//FILE* fp = fopen("haha", "a");
//fprintf(fp, "boo! %x %x %d %s %s 0%o\n", ip_, someInterp, recursive, fn_, mode_, perm_);
//fclose(fp);
    if (ip_ == 0)
    {
        ip_ = someInterp;
	if (ip_ == 0)
	{
	    // there's no way to reflect to Tcl, so this has to fail
	    Tcl_SetErrno(EINVAL);
	    return 0;
	}
    }

    Tcl_Channel result = 0;
    Tcl_MutexLock(&icMutex);
    if (!recursive)
    {
	recursive = 1;
	Tcl_Obj* a = Tcl_NewListObj(0, 0);
	Tcl_IncrRefCount(a);

	int e = Tcl_ListObjAppendElement(ip_, a,
				Tcl_NewStringObj("vfs::open", -1));
	if (e == TCL_OK)
	    e = Tcl_ListObjAppendElement(ip_, a, Tcl_NewStringObj(fn_, -1));
	if (e == TCL_OK)
	    e = Tcl_ListObjAppendElement(ip_, a, Tcl_NewStringObj(mode_, -1));
	if (e == TCL_OK)
	{
	    Tcl_SavedResult sr;
	    Tcl_SaveResult(ip_, &sr);
            e = Tcl_EvalObjEx(ip_, a, TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
            if (e == TCL_OK)
	    {
            	result = Tcl_GetChannel(ip_, Tcl_GetStringResult(ip_), 0);
		if (result != 0)
		{
		    // Changed between 8.4a1 and 8.4a2, choose accordingly
		    //((Channel*) result)->refCount++;
		    ((Channel*) result)->state->refCount++;
		    e = Tcl_UnregisterChannel(ip_, result);
		    //((Channel*) result)->refCount--;
		    ((Channel*) result)->state->refCount--;
		    if (e != TCL_OK)
			result = 0;
		}
	    }
	    //printf("result %d: %s\n", e, Tcl_GetStringResult(ip_));
	    Tcl_RestoreResult(ip_, &sr);
	}

	Tcl_DecrRefCount(a);
	recursive = 0;
    }
    Tcl_MutexUnlock(&icMutex);
    return result;
}

static int
VfsChanObjCmd(ClientData cd_, Tcl_Interp* ip_, int objc_, Tcl_Obj*const* objv_)
{
    if (someInterp == 0 && ip_ != 0)
	someInterp = ip_;
    TclOpenFileChannelInsertProc(VfsChanFileOpen);
    
    return TCL_OK;
}

///////////////////////////////////////////////////////////////////////////////

DLLEXPORT int Rechan_Init(Tcl_Interp* interp)
{
    if (!Tcl_InitStubs(interp, "8.2", 0))
        return TCL_ERROR;
    Tcl_CreateObjCommand(interp, "rechan", RechanObjCmd, 0, 0);
    Tcl_CreateObjCommand(interp, "vfschan", VfsChanObjCmd, 0, 0);
    return Tcl_PkgProvide(interp, "rechan", "0.2");
}
