// Copyright (c) 1999-2000 Jean-Claude Wippler <jcw@equi4.com>
//
// C++ Tcl command extension

#include "wcommon.h"
#include "wtclimpl.h"
#include "wutil.h"

#include <string.h>

/////////////////////////////////////////////////////////////////////////////

wRef::wRef (wObjImpl* ptr_)
    : _ptr (ptr_) 
{ 
    if (_ptr == 0)
        _ptr = (wObjImpl*) Tcl_NewObj();
    Tcl_IncrRefCount(&_ptr->_o); 
}

wRef::~wRef () 
{ 
    Tcl_DecrRefCount(&_ptr->_o); 
}

wRef::wRef (const wRef& obj_) 
    : _ptr (obj_._ptr) 
{ 
    Tcl_IncrRefCount(&_ptr->_o); 
}

void wRef::operator= (const wRef& obj_)
{ 
    Tcl_IncrRefCount(&obj_._ptr->_o);
    Tcl_DecrRefCount(&_ptr->_o);
    _ptr = obj_._ptr;
}

wRef::operator wObjImpl* () const
{
    return _ptr;
}

wRef::wRef (int value_)
    : _ptr ((wObjImpl*) Tcl_NewIntObj(value_))
{
    A(_ptr != 0);
    Tcl_IncrRefCount(&_ptr->_o); 
}

bool wRef::IsInt() const
{
    int result;
    return Tcl_GetIntFromObj(0, &_ptr->_o, &result) == TCL_OK;
}

wRef::operator int () const
{
    int result;
    int e = Tcl_GetIntFromObj(0, &_ptr->_o, &result);
    if (e != TCL_OK)
        throw e;
    return result;
}

wRef::wRef (long value_)
    : _ptr ((wObjImpl*) Tcl_NewLongObj(value_))
{
    A(_ptr != 0);
    Tcl_IncrRefCount(&_ptr->_o); 
}

bool wRef::IsLong() const
{
    long result;
    return Tcl_GetLongFromObj(0, &_ptr->_o, &result) == TCL_OK;
}

wRef::operator long () const
{
    long result;
    int e = Tcl_GetLongFromObj(0, &_ptr->_o, &result);
    if (e != TCL_OK)
        throw e;
    return result;
}

wRef::wRef (double value_)
    : _ptr ((wObjImpl*) Tcl_NewDoubleObj(value_))
{
    A(_ptr != 0);
    Tcl_IncrRefCount(&_ptr->_o); 
}

bool wRef::IsDouble() const
{
    double result;
    return Tcl_GetDoubleFromObj(0, &_ptr->_o, &result) == TCL_OK;
}

wRef::operator double () const
{
    double result;
    int e = Tcl_GetDoubleFromObj(0, &_ptr->_o, &result);
    if (e != TCL_OK)
        throw e;
    return result;
}

wRef::wRef (const char* value_)
    : _ptr ((wObjImpl*) Tcl_NewStringObj(value_, -1))
{
    A(_ptr != 0);
    Tcl_IncrRefCount(&_ptr->_o); 
}

wRef::wRef (const void* value_, int length_)
    : _ptr ((wObjImpl*) Tcl_NewByteArrayObj((unsigned char*) value_, length_))
{
    A(_ptr != 0);
    Tcl_IncrRefCount(&_ptr->_o); 
}

bool wRef::IsString() const
{
    return true;
}

int wRef::Length() const
{
    if (_ptr->_o.bytes != 0) // optimization
        return _ptr->_o.length;
    
    int result;
    (void) Tcl_GetStringFromObj(&_ptr->_o, &result);
    return result;
}

int wRef::ByteLength() const
{
    int result;
    (void) Tcl_GetByteArrayFromObj(&_ptr->_o, &result);
    return result;
}

wRef::operator const char* () const
{
    return Tcl_GetStringFromObj(&_ptr->_o, 0);
}

wRef::operator const void* () const
{
    return Tcl_GetByteArrayFromObj(&_ptr->_o, 0);
}

bool wRef::IsSeq() const
{
    int result;
    return Tcl_ListObjLength(0, &_ptr->_o, &result) == TCL_OK;
}

wObject& wRef::Cast(const char* type_) const
{
    A(false); // not yet
    return *(wObject*) 0;
}

/////////////////////////////////////////////////////////////////////////////

    static wObjImpl* Own(wObjImpl*& ptr_)
    {
        if (Tcl_IsShared(&ptr_->_o))
            ptr_ = (wObjImpl*) Tcl_DuplicateObj(&ptr_->_o);
        return ptr_;
    }

wSeqRef::wSeqRef ()
    : wRef ((wObjImpl*) Tcl_NewListObj(0, 0))
{
}

wSeqRef::wSeqRef (const wRef& obj_)
    : wRef (obj_)
{
    int result;
    int e = Tcl_ListObjLength(0, &((wObjImpl*) obj_)->_o, &result);
    if (e != TCL_OK)
        throw e;
}

void wSeqRef::operator= (const wRef& obj_)
{
    int result;
    int e = Tcl_ListObjLength(0, &((wObjImpl*) obj_)->_o, &result);
    if (e != TCL_OK)
        throw e;

    wRef::operator= (obj_);
}

int wSeqRef::Size() const
{
    int result;
    V(Tcl_ListObjLength(0, &_ptr->_o, &result) == TCL_OK);
    return result;
}

wObjImpl* wSeqRef::operator[] (int index_) const
{
    wObjImpl* result;
    V(Tcl_ListObjIndex(0, &_ptr->_o, index_, (Tcl_Obj**) &result) == TCL_OK);
    return result;
}

void wSeqRef::SetAt(int index_, const wRef& obj_)
{
    wObjImpl* value = obj_;
    V(Tcl_ListObjReplace(0, &Own(_ptr)->_o, index_, 1, 1, (Tcl_Obj**) &value) == TCL_OK);
}

    static void wFixRange(int low_, int& begin_, int& end_, int high_)
    {
        if (begin_ < low_)
            begin_ = low_;
        if (end_ > high_)
            end_ = high_;
        if (end_ < begin_)
            end_ = begin_;
    }

wSeqRef wSeqRef::GetSlice(int begin_, int end_) const
{
    int count;
    wObjImpl** list;
    V(Tcl_ListObjGetElements(0, &_ptr->_o, &count, (Tcl_Obj***) &list) == TCL_OK);

    wFixRange(0, begin_, end_, count);
    return (wRef) (wObjImpl*) Tcl_NewListObj(end_ - begin_, (Tcl_Obj**) list + begin_);
}

void wSeqRef::SetSlice(int begin_, int end_, const wSeqRef& list_)
{
    wFixRange(0, begin_, end_, Size());

    int count;
    wObjImpl** list;
    V(Tcl_ListObjGetElements(0, (Tcl_Obj*)(const void*) list_, &count, (Tcl_Obj***)(void*) &list) == TCL_OK);

    V(Tcl_ListObjReplace(0, &Own(_ptr)->_o, begin_, end_ - begin_,
                                count, (Tcl_Obj**) list) == TCL_OK);
}

void wSeqRef::Append(const wRef& obj_)
{
    V(Tcl_ListObjAppendElement(0, &Own(_ptr)->_o, &((wObjImpl*) obj_)->_o) == TCL_OK);
}

void wSeqRef::Insert(int index_, const wRef& obj_, int count_)
{
    wObjImpl* value = obj_;

    if (count_ == 1)
    {
        V(Tcl_ListObjReplace(0, &Own(_ptr)->_o, index_, 0, 1, (Tcl_Obj**) &value) == TCL_OK);
    }
    else if (count_ > 1) 
    {
        wObjImpl** vec = new wObjImpl* [count_];
        for (int i = 0; i < count_; ++i)
            vec[i] = value;
        V(Tcl_ListObjReplace(0, &Own(_ptr)->_o, index_, 0, count_, (Tcl_Obj**) vec) == TCL_OK);
        delete[] vec;
    }
}

void wSeqRef::Delete(int index_, int count_)
{
    V(Tcl_ListObjReplace(0, &Own(_ptr)->_o, index_, count_, 0, 0) == TCL_OK);
}

/////////////////////////////////////////////////////////////////////////////

wCallFrame::wCallFrame (wContext* context_, int objc_, wObjImpl** objv_)
    : _context (context_), _objc (objc_), _objv (objv_), _pos (1)
{
}

int wCallFrame::Lookup(const char** cmds_)
{
    if (_pos >= _objc)
        throw TCL_ERROR;

    int result;
    int e = Tcl_GetIndexFromObj(&_context->_o, &_objv[_pos++]->_o, 
                    (char**) cmds_, "option", 0, &result);
    if (e != TCL_OK)
        throw e;
    return result;
}

wRef wCallFrame::RegisterCmd(wObject* obj_)
{
    wRef result = (long) this;
    if (!obj_->Register(_context, result))
        throw TCL_ERROR;
    return result;
}

bool wCallFrame::More() const
{
    return _pos < _objc;
}

wRef wCallFrame::operator++ ()
{
    if (!More())
        throw TCL_ERROR;

    return _objv[_pos++];
}

wSeqRef wCallFrame::Tail()
{
    int pos = _pos;
    _pos = _objc;
    return (wRef) (wObjImpl*) Tcl_NewListObj(_objc - pos, (Tcl_Obj**) _objv + pos);
}

void Done(wCallFrame& sf_)
{
    if (sf_.More())
        throw TCL_ERROR;
}

wCallFrame& wCallFrame::operator>> (int& arg_)
{
    if (More())
        arg_ = ++*this;
    return *this;
}

wCallFrame& wCallFrame::operator>> (long& arg_)
{
    if (More())
        arg_ = ++*this;
    return *this;
}

wCallFrame& wCallFrame::operator>> (double& arg_)
{
    if (More())
        arg_ = ++*this;
    return *this;
}

wCallFrame& wCallFrame::operator>> (const char*& arg_)
{
    if (More())
        arg_ = ++*this;
    return *this;
}

wCallFrame& wCallFrame::operator>> (wRef& arg_)
{
    if (More())
        arg_ = ++*this;
    return *this;
}

void wCallFrame::operator>> (void (*fun_)(wCallFrame&))
{
    fun_(*this);
}

wCallFrame::operator wContext* () const
{
    return _context;
}

/////////////////////////////////////////////////////////////////////////////

    static int wObjectCmdProc(ClientData self_, wContext* context_,
                                int objc_, wObjImpl** objv_)
    {   
        wObject* self = (wObject*) self_;
        A(self != 0 && context_ != 0);
        A(context_ == self->_context);

        wCallFrame cf (self->_context, objc_, objv_);

        try
        {
            Tcl_SetObjResult(&self->_context->_o,
		    		&((wObjImpl*) self->Call(cf))->_o);
            Done(cf);
        }
        catch (int e_)
        {
	    (void) e_; // e_ is only there so a debugger can inspect it
            A(false);
            return TCL_ERROR;
        }
        catch (const char* e_)
        {
	    Tcl_SetObjResult(&self->_context->_o,
		    		Tcl_NewStringObj(e_,-1));
            A(false);
            return TCL_ERROR;
        }
        catch (...)
        {
            A(false);
            return TCL_ERROR;
        }

        return TCL_OK;
    }

    static void wObjectDelProc(ClientData self_)
    {
        wObject* self = (wObject*) self_;
        A(self != 0);
        delete self;
    }

wObjectBase::wObjectBase ()
    : _token (0), _context (0)
{
}

wObject::~wObject ()
{
    if (_token != 0) {
        Tcl_DeleteCommandFromToken(&_context->_o, (Tcl_Command) _token);
    }
}

wRef wObject::Call(wCallFrame& cf_)
{
    A(false);
    return wRef ();
}

bool wObject::Register(wContext* context_, const char* name_) 
{
    A(_token == 0);
    _context = context_;
    _token = Tcl_CreateObjCommand(&context_->_o, (char*) name_,
                (Tcl_ObjCmdProc*) wObjectCmdProc, this, wObjectDelProc);
    return _token != 0;
}

/////////////////////////////////////////////////////////////////////////////

class wPackage : public wObject
{
    wRef (*_proc)(wCallFrame&);

    static void ExitProc(ClientData);

public:
    wPackage (wRef (*proc_)(wCallFrame&));
    virtual ~wPackage ();

    virtual wRef Call(wCallFrame& cf_);
    bool Register(wContext* context_, const char* name_);
};

wPackage::wPackage (wRef (*proc_)(wCallFrame&))
    : _proc (proc_)
{
}

wPackage::~wPackage ()
{
    Tcl_DeleteExitHandler(wPackage::ExitProc, this);
}

wRef wPackage::Call(wCallFrame& cf_)
{
    A(_proc != 0);
    return _proc(cf_);
}

bool wPackage::Register(wContext* context_, const char* name_)
{
    bool ok = wObject::Register(context_, name_);
    if (ok)
        Tcl_CreateExitHandler(wPackage::ExitProc, this);
    return ok;
}

void wPackage::ExitProc(ClientData self_)
{
    wPackage* self = (wPackage*) self_;
    A(self != 0);
    Tcl_DeleteCommandFromToken(&self->_context->_o, (Tcl_Command) self->_token);
}

/////////////////////////////////////////////////////////////////////////////

int wDefine(wRef (*proc_)(wCallFrame&), const char* name_,
                    wContext* context_, const char* version_)
{
    A(context_ != 0);

    #if USE_TCL_STUBS
        if (Tcl_InitStubs(&context_->_o, "8.1", 0) == NULL) 
            return TCL_ERROR;
    #endif

    wPackage* pkg = new wPackage (proc_);
    if (!pkg->Register(context_, name_))
    {
        delete pkg;
        return TCL_ERROR;
    }

    return Tcl_PkgProvide (&context_->_o, (char*) name_, (char*) version_);
}

/////////////////////////////////////////////////////////////////////////////
